#include "sim.h"
#include "server.h"

/* mime_extension
 * Returns the mime type corresponding to the extension on the given filename
 */
char *mime_extension(char *fn)
{
static char typebuf[256];	/* buffer for mime type */
char *extp,ext[256];		/* file extension */
int fd = -1;			/* fd for reading .mime.types file */

/* Determine extension
 */
for(extp=fn+strlen(fn)+1; *extp != '.'; extp--)
	if (extp < fn)
		return "application/octet-stream";
strcpy(ext,extp+1);
for(extp=ext; *extp; extp++)
	if (isupper(*extp)) *extp = tolower(*extp);

/* Try to open .mime.types file
 */
if (getenv("HOME")) {
	char mimetypes[256];	/* filename for .mime.types file */

	sprintf(mimetypes,"%s/%s",getenv("HOME"),MIME_TYPES);
	fd = open(mimetypes,O_RDONLY);
	}
if (fd == -1) {
	fd = open(MIME_TYPES,O_RDONLY);
	}
if (fd != -1) {
	/* Lookup extension in file
	 */
	char line[256];

	while(fdgets(fd,line,256) != NULL) {
		char *type;
		char *ex;

		if ((type = strtok(line," \t\n")) == NULL)
			continue;
		while((ex = strtok(NULL," \t\n"))) {
			if (strequ(ex,ext)) {
				strcpy(typebuf,type);
				close(fd);
				return typebuf;
				}
			}
		}
	close(fd);
	}

/* Didn't find type. Use internal definitions
 */
if (strequ(ext,"ps"))	return "application/postscript";
if (strequ(ext,"tex"))	return "application/x-tex";
if (strequ(ext,"au"))	return "audio/basic";
if (strequ(ext,"aifc"))	return "audio/x-aiff";
if (strequ(ext,"aiff"))	return "audio/x-aiff";
if (strequ(ext,"wav"))	return "audio/x-wav";
if (strequ(ext,"mod"))  return "audio/x-mod";
if (strequ(ext,"gif"))	return "image/gif";
if (strequ(ext,"jpg"))	return "image/jpeg";
if (strequ(ext,"jpeg"))	return "image/jpeg";
if (strequ(ext,"tif"))	return "image/tiff";
if (strequ(ext,"tiff"))	return "image/tiff";
if (strequ(ext,"ppm"))	return "image/ppm";
if (strequ(ext,"pbm"))	return "image/pbm";
if (strequ(ext,"html"))	return "text/html";
if (strequ(ext,"htm"))	return "text/html";
if (strequ(ext,"txt"))	return "text/plain";
if (strequ(ext,"c"))	return "text/plain";
if (strequ(ext,"h"))	return "text/plain";
if (strequ(ext,"cc"))	return "text/plain";
if (strequ(ext,"mpeg"))	return "video/mpeg";
if (strequ(ext,"mpg"))	return "video/mpeg";
if (strequ(ext,"qt"))	return "video/quicktime";
if (strequ(ext,"mov"))	return "video/quicktime";
if (strequ(ext,"avi"))	return "video/x-msvideo";
if (strequ(ext,"mv"))	return "video/x-sgi-movie";
if (strequ(ext,"movie"))return "video/x-sgi-movie";
if (strequ(ext,"url"))	return "text/x-url";
if (strequ(ext,"tel"))	return "text/x-tel";
if (strequ(ext,"shape"))return "text/plain";
return "application/octet-stream";
}


/* send_mime_data
 * Concurrently send a named file to a client.
 */
int send_mime_data(int fd, char *file)
{
int f;					/* fd for reading from file */
struct stat stbuf;			/* info on file to send */
int port = mp->port + alloc_port();	/* port client connects to get data */

if ((f = open(file,O_RDONLY)) == -1)
	return 0;
fstat(f,&stbuf);
server_command(fd,com_displaymime,mime_extension(basenam(file)),
	       stbuf.st_size,port);
if (!sfork()) {
	set_signals(SIG_DFL);
	if (send_data(port,f,stbuf.st_size) == 0)
		logerror("Error sending file %s to client",file);
	close(f);
	exit(0);
	}
close(f);
return 1;
}


/* send_mime_string
 * Concurrently send a string as mime data to a client.
 */
void send_mime_string(int fd, char *dat, char *type)
{
int len = strlen(dat);			/* length of string */
char fname[256];			/* fake filename for string */
int port = mp->port + alloc_port();	/* port client connects to get data */
char *tn = tmpnam(NULL);		/* temp file for data */

sprintf(fname,".%s",type);
server_command(fd,com_displaymime,mime_extension(fname),len,port);
if (!sfork()) {
	/* Concurrently send data to client
	 */
	int f;		/* fd for storing string in temp file */

	set_signals(SIG_DFL);
	if ((f = creat(tn,0700)) == -1) {
		logerror("Couldn't create temp file!?");
		exit(0);
		}
	fdwrite(f,dat,1,len);
	close(f);
	if ((f = open(tn,O_RDONLY)) == -1) {
		logerror("Could open temp file!!!?");
		exit(0);
		}
	send_data(port,f,len);
	close(f);
	unlink(tn);
	exit(0);
	}
}


/* send_url_data
 * Gets data from a given url, and sends it to the client on the given fd
 */
int send_url_data(int pfd, char *url)
{
char buf[1024];
char *method, *host, *port, *dir;	/* for URL */
int fd;

strcpy(buf,url);
if ((method = strtok(buf,":")) == NULL) {
	logerror("Missing access method for URL");
	return 0;
	}
if (strequ(method,"http")) {
	/* Get data from a web server
	 */
	char statline[256];	/* first line from web server */
	char *protocol, *code;	/* from statline */
	char content_type[256] = "application/octet-stream";
	int content_length = 0;
	int cport;		/* port for transfer to client */

	if ((host = strtok(NULL,"/")) == NULL) {
		logerror("Missing host name for http URL");
		return 0;
		}
	dir = host+strlen(host)+1;
	if ((port = strchr(host,':'))) {
		*port = 0;
		port++;
		}
	else
		port = "80";

	/* Connect and send request
	 */
	fd = open_tcp_send(host,atoi(port));
	if (fd == -1) {
		logerror("Couldn't connect to %s",host);
		return 0;
		}
	fdprintf(fd,"GET /%s HTTP/1.0\n",dir);
	fdprintf(fd,"User-Agent: SimServer V%s\n",SIM_VERSION);
	fdputc(fd,'\n');

	/* Read HTTP headers
	 */
	fdgets(fd,statline,256);
	protocol = strtok(statline," ");
	if (!protocol || !strequ(protocol,"HTTP/1.0")) {
		logerror("Unknown protocol %s",protocol);
		return 0;
		}
	code = strtok(NULL," ");
	if (!code) {
		logerror("No status code?!");
		return 0;
		}
	else if (!strequ(code,"200") && !strequ(code,"301") &&
		 !strequ(code,"302")) {
			logerror("Request failed, status %s",code);
			return 0;
			}
	while(1) {
		char headline[256];	/* header line from web server */
		char *entity, *value;
		char *lastch;

		/* Get entity and value strings
		 */
		fdgets(fd,headline,256);
		lastch = headline+strlen(headline)-1;
		while(1) {
			if (*lastch==' ' || *lastch=='\r' || *lastch=='\n')
				*(lastch--) = 0;
			else
				break;
			}
		if (!strlen(headline) || !(value = strchr(headline,':')))
			break;
		*value = 0;
		while(*(++value) == ' ')
			;
		entity = headline;

		/* Parse entity
		 */
		if (strequ(entity,"Content-type"))
			strcpy(content_type,value);
		else if (strequ(entity,"Content-length"))
			content_length = atoi(value);
		else if (strequ(entity,"Location")) {
			/* Totally new URL!
			 */
			return send_url_data(pfd,value);
			}
		}

	/* Send data to player
	 */
	if (!content_length) {
		logerror("Didn't get a Content-length header");
		return 0;
		}
	cport = mp->port + alloc_port();
	server_command(pfd,com_displaymime,content_type,content_length,cport);
	if (!sfork()) {
		set_signals(SIG_DFL);
		if (!send_data(cport,fd,content_length))
			logerror("Error sending data to client");
		close(fd);
		exit(0);
		}
	close(fd);
	}
else {
	logerror("Unknown access method for URL");
	return 0;
	}
return 1;
}

